
'''
    - Classes for TraceWin elements/commands for external manipulations.
    - First double check all the used elements/commands.

    - Element/command types

      * Active elements

        FIELD_MAP

        DRIFT
        QUAD
        THIN_STEERING
        GAP
        DTL_CEL

      * Passive elements

        BEND
        EDGE
        APERTURE
        DIAG_POSITION
        DIAG_...

      * Active commands

        STEERER
        CHOPPER

      * Passive commands

        FREQ
        MARKER

      * Other commands for just common "COMM"

        END
        LATTICE
        LATTICE_END

        SET_ADV
        SET_SIZE
        SET_TWISS
        SET_ACHROMAT
        MATCH_FAM_GRAD
        MATCH_FAM_FIELD
        MATCH_FAM_PHASE
        MATCH_FAM_LFOC
        MIN_PHASE_VARIATION
        START_ACHROMAT

        ERROR_QUAD_NCPL_STAT
        ERROR_QUAD_CPL_STAT
        ERROR_CAV_NCPL_STAT
        ERROR_CAV_NCPL_DYN
        ERROR_CAV_CPL_STAT
        ERROR_CAV_CPL_DYN
        ADJUST
        ADJUST_STEERER
        ADJUST_STEERER_BX
        ADJUST_STEERER_BY
        SET_BEAM_PHASE_ERROR

        PLOT_DST

    2016.01.19
'''

#---- Libs

from numpy import *
from sys   import exit

#---- Constants

#mass=938.272046  # Wiki
mass=938.272029  # TraceWin
c   =2.99792458

#---- Misc functions/classes

def Brho(gamma):

    return (10.0/c)*sqrt(gamma**2-1.0)*mass*1e-3

class FIELD_MAP_DATA:
    '''
        - Only 1D (Ez) supported for now.
    '''
    def __init__(self,file_name):

        # Read file
        with open(file_name) as file: lin=file.readlines()

        # Instances
        self.dL   =float(lin[0].split()[1])/int(lin[0].split()[0])
        self.field=map(float,lin[2:])

#---- Parent classes

class ELEM:
    '''
    '''
    def __init__(self,name,typ,para):

        # Basic instances
        self.name=name
        self.typ =typ
        self.para=para

        # Option instances
        self.L       = 0.0
        self.apt     = None
        self.idx     =-1
        self.idx_elem=-1
        self.s       = 0.0
        self.gamma   = 1.0
        self.freq    = 352.21

    def update(self):
        '''
          Updates element based on para
        '''
        pass

    def update_idx(self):

        # Update idx, idx_elem, s
        self.idx     +=1
        self.idx_elem+=1
        self.s       +=self.L

    def get_tw(self):

        if self.name!='': lin=self.name+': '+self.typ+' '+' '.join(self.para)
        if self.name=='': lin=               self.typ+' '+' '.join(self.para)

        return lin

    ## def get_madx(self):

    ##     i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

    ##     if self.name!='': lin= self.name        +': marker;'
    ##     if self.name=='': lin='elem'+i+'_marker'+': marker;'

    ##     return lin

class COMM:
    '''
        - Same inputs as ELEM class but some may not be needed ...
    '''
    def __init__(self,name,typ,para):

        # Basic instances
        self.name=name
        self.typ =typ
        self.para=para

        # Option instances
        self.idx     =-1
        self.idx_elem=-1
        self.s       = 0.0
        self.gamma   = 1.0
        self.freq    = 352.21

    def __repr__(self):
        S="lib_tw_elem.COMM: " + self.name + ", type: " + self.typ + ", para: " + str(self.para)
        S+="\n  idx: " + str(self.idx) + ", idx_elem: " + str(self.idx_elem)
        S+="\n  s: " + str(self.s) + ", gamma:" + str(self.gamma)
        S+="\n  freq: " + str(self.freq)
        return S

    def update_idx(self):

        # Update idx
        self.idx+=1

    def get_tw(self):

        if self.name!='': lin=self.name+': '+self.typ+' '+' '.join(self.para)
        if self.name=='': lin=               self.typ+' '+' '.join(self.para)

        return lin

#---- Field map (need an additional input, dic_fmap)

class FIELD_MAP(ELEM):
    '''
        - Calculation of gamma only 1D (Ez) supported.
    '''
    def __init__(self,name,typ,para,dic_fmap):

        ELEM.__init__(self,name,typ,para)
        # TW instances
        self.name_fmap=      para[8]
        self.typ_fmap =      para[0]
        self.L        =float(para[1])*1e-3      # mm => m
        self.phs_rf   =float(para[2])*pi/180.0  # deg => rad
        self.apt      =float(para[3])
        self.Bnom     =float(para[4])
        self.Enom     =float(para[5])
        self.sc_comp  =float(para[6])
        self.flag_apt =      para[7]

        # Option instances
        self.data=dic_fmap[self.name_fmap]

    def update_gamma(self):

        # Temp error message for field map types
        if self.typ_fmap!='100': print 'Gamma calc only supported for 1D, exiting ...', exit()

        # Update gamma (simple ver, closer to TW)
        dL=self.data.dL; V=self.Enom*array(self.data.field)*dL; phs=self.phs_rf; C=0.0; S=0.0
        for i in range(len(V)):
            C+=V[i]*cos(phs); S+=V[i]*sin(phs); self.gamma+=V[i]/mass*cos(phs)
            beta=sqrt(1.0-1.0/self.gamma**2); phs+=2.0*pi*self.freq*dL/(beta*c*1e2)

        # Update gamma (I think this is more accurate...)
        ## dL=self.data.dL; V=self.Enom*array(self.data.field)*dL; phs=self.phs_rf; C=0; S=0
        ## for i in range(len(V)):
        ##     if i==0        :
        ##         C+=V[i]*cos(phs); S+=V[i]*sin(phs); self.gamma+=0.5*V[i]/mass*cos(phs)
        ##         beta=sqrt(1.0-1.0/self.gamma**2); phs+=2.0*pi*self.freq*dL/(beta*c*1e2)
        ##     if 0<i<len(V)-1:
        ##         C+=V[i]*cos(phs); S+=V[i]*sin(phs); self.gamma+=    V[i]/mass*cos(phs)
        ##         beta=sqrt(1.0-1.0/self.gamma**2); phs+=2.0*pi*self.freq*dL/(beta*c*1e2)
        ##     if i==len(V)-1 :
        ##         C+=V[i]*cos(phs); S+=V[i]*sin(phs); self.gamma+=0.5*V[i]/mass*cos(phs)

        # Option instances for other codes (MADX?)
        self.phs_syn=arctan(S/C)
        self.E0TL   =C/cos(self.phs_syn)

    def get_tw(self):

        if self.name!='': lin=self.name+': '+self.typ+' '
        if self.name=='': lin=               self.typ+' '
        lin+=    self.typ_fmap        +' '
        lin+=str(self.L*1e3          )+' '
        lin+=str(self.phs_rf*180.0/pi)+' '
        lin+=str(self.apt            )+' '
        lin+=str(self.Bnom           )+' '
        lin+=str(self.Enom           )+' '
        lin+=str(self.sc_comp        )+' '
        lin+=    self.flag_apt        +' '
        lin+=    self.name_fmap

        return lin

    def get_madx(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name          +': RFCAVITY, '
        if self.name=='': lin='ELEM'+i+'_FIELDMAP'+': RFCAVITY, '
        lin+='L='   +str(self.L                        )+', '
        lin+='FREQ='+str(self.freq                     )+', '
        lin+='VOLT='+str(self.E0TL                     )+', '
        lin+='LAG=' +str((0.5*pi-self.phs_syn)/(2.0*pi))+'; '  # phs_MADX = pi/2-phs_TW

        return lin

    def get_fluka(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name          +'  RFCAVITY  '
        if self.name=='': lin='ELEM'+i+'_FIELDMAP'+'  RFCAVITY  '
        lin+=str(self.L)+'  '+str(self.s)+'  '
        lin+='0.0  0.0  0.0  '
        lin+='CIRCLE  '+str(self.apt*1e-3)+'  0.0'

        return lin

    def get_mars(self):

        lin ='"RFCAVITY"  ""  ""  '+str(self.s)+'  '+str(self.L)+'  '
        lin+='0  0  0  0  0'

        return lin

#---- Active elements

class DRIFT(ELEM):
    '''
    '''
    def __init__(self,name,typ,para):

        ELEM.__init__(self,name,typ,para)

        # TW instances
        self.L  =float(para[0])*1e-3  # mm => m
        self.apt=float(para[1])

        # TW option instances
        try   : self.apty=float(para[2])
        except: self.apty=0.0

    def get_tw(self):

        if self.name!='': lin=self.name+': '+self.typ+' '
        if self.name=='': lin=               self.typ+' '
        lin+=str(self.L*1e3)+' '
        lin+=str(self.apt  )+' '
        lin+=str(self.apty )

        return lin

    def get_madx(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name       +': DRIFT, L='+str(self.L)+';'
        if self.name=='': lin='ELEM'+i+'_DRIFT'+': DRIFT, L='+str(self.L)+';'

        return lin

    def get_fluka(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name       +'  DRIFT  '
        if self.name=='': lin='ELEM'+i+'_DRIFT'+'  DRIFT  '
        lin+=str(self.L)+'  '+str(self.s)+'  '
        lin+='0.0  0.0  0.0  '
        if self.apty==0.0: lin+='CIRCLE  '+str(self.apt*1e-3)+'  0.0'
        if self.apty!=0.0: lin+='RECTANGLE  '+str(self.apt*1e-3)+'  '+str(self.apty*1e-3)

        return lin

    def get_mars(self):

        lin ='"DRIFT"  ""  ""  '+str(self.s)+'  '+str(self.L)+'  '
        lin+='0  0  0  0  0'

        return lin

class QUAD(ELEM):
    '''
        - Tilt not supported.
    '''
    def __init__(self,name,typ,para):

        ELEM.__init__(self,name,typ,para)

        # TW instances
        self.update()
        self.tilt=0.0

        # TW option instances
        try   : self.b3    =float(para[4])
        except: self.b3    =0.0
        try   : self.b4    =float(para[5])
        except: self.b4    =0.0
        try   : self.b5    =float(para[6])
        except: self.b5    =0.0
        try   : self.b6    =float(para[7])
        except: self.b6    =0.0
        try   : self.r_good=float(para[8])
        except: self.r_good=0.0

    def update(self):
        '''
          Updates element based on para
        '''
        self.L   =float(self.para[0])*1e-3  # mm => m
        self.G   =float(self.para[1])
        self.apt =float(self.para[2])

    def get_tw(self):

        if self.name!='': lin=self.name+': '+self.typ+' '
        if self.name=='': lin=               self.typ+' '
        lin+=str(self.L*1e3 )+' '
        lin+=str(self.G     )+' '
        lin+=str(self.apt   )+' '
        lin+=str(self.tilt  )+' '
        lin+=str(self.b3    )+' '
        lin+=str(self.b4    )+' '
        lin+=str(self.b5    )+' '
        lin+=str(self.b6    )+' '
        lin+=str(self.r_good)

        return lin

    def get_madx(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name      +': QUADRUPOLE, '
        if self.name=='': lin='ELEM'+i+'_QUAD'+': QUADRUPOLE, '
        lin+='L=' +str(self.L                 )+', '
        lin+='K1='+str(self.G/Brho(self.gamma))+'; '

        return lin

    def get_fluka(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name      +'  QUADRUPOLE  '
        if self.name=='': lin='ELEM'+i+'_QUAD'+'  QUADRUPOLE  '
        lin+=str(self.L)+'  '+str(self.s)+'  '
        lin+='0.0  0.0  '+str(self.G*self.L/Brho(self.gamma))+'  '
        lin+='CIRCLE  '+str(self.apt*1e-3)+'  0.0'

        return lin

    def get_mars(self):

        lin ='"QUADRUPOLE"  ""  ""  '+str(self.s)+'  '+str(self.L)+'  '
        lin+='0  '+str(self.G/Brho(self.gamma))+'  0  0  0  '

        return lin

class THIN_STEERING(ELEM):
    '''
        - Only magnetic steerer supported.
        - Info of plane missing.
    '''
    def __init__(self,name,typ,para):

        ELEM.__init__(self,name,typ,para)

        # TW instances
        self.update()
        self.typ_EM='0'

    def update(self):
        '''
          Updates element based on para
        '''
        self.BLx   =float(self.para[0])
        self.BLy   =float(self.para[1])
        self.apt   =float(self.para[2])

    def get_tw(self):

        if self.name!='': lin=self.name+': '+self.typ+' '
        if self.name=='': lin=               self.typ+' '
        lin+=str(self.BLx   )+' '
        lin+=str(self.BLy   )+' '
        lin+=str(self.apt   )+' '
        lin+=    self.typ_EM

        return lin

    def get_madx(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name         +': KICKER, L=0, '
        if self.name=='': lin='ELEM'+i+'_STEERER'+': KICKER, L=0, '
        lin+='HKICK='+str(-self.BLy/Brho(self.gamma))+', '
        lin+='VKICK='+str( self.BLx/Brho(self.gamma))+'; '

        return lin

    def get_fluka(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name         +'  KICKER  '
        if self.name=='': lin='ELEM'+i+'_STEERER'+'  KICKER  '
        lin+=str(self.L)+'  '+str(self.s)+'  '
        lin+='0.0  0.0  0.0  '
        lin+='CIRCLE  '+str(self.apt*1e-3)+'  0.0'

        return lin

    def get_mars(self):

        lin ='"KICKER"  ""  ""  '+str(self.s)+'  '+str(self.L)+'  '
        lin+='0  0  0  0  0'

        return lin

class GAP(ELEM):
    '''
        - Note there are E0T0L (for TW) and E0TL (for other codes).
    '''
    def __init__(self,name,typ,para):

        ELEM.__init__(self,name,typ,para)

        # TW instances
        self.E0T0L     =float(para[0])*1e-6      # V => MV
        self.phs_rf    =float(para[1])*pi/180.0  # deg => rad
        self.apt       =float(para[2])
        self.phs_rf_typ=  int(para[3])

        # TW option instances
        try   : self.beta_s=float(para[4])
        except: self.beta_s=0.0
        try   : self.T0    =float(para[5])
        except: self.T0    =0.0
        try   : self.T1    =float(para[6])
        except: self.T1    =0.0
        try   : self.T2    =float(para[7])
        except: self.T2    =0.0

        # Option instances
        self.E0TL=self.E0T0L

    def update_gamma(self):

        # Update gamma (and E0TL)
        if self.beta_s!=0.0 and self.T0!=0.0:
            gamma_c   =self.gamma+0.5*self.E0TL/mass*cos(self.phs_rf)
            beta_c    =sqrt(1.0-1.0/gamma_c**2)
            k         =self.beta_s/beta_c
            self.E0TL*=(self.T0+self.T1*(k-1.0)+0.5*self.T2*(k-1.0)**2)/self.T0
        self.gamma+=self.E0TL/mass*cos(self.phs_rf)

    def get_tw(self):

        if self.name!='': lin=self.name+': '+self.typ+' '
        if self.name=='': lin=               self.typ+' '
        lin+=str(self.E0T0L*1e6      )+' '
        lin+=str(self.phs_rf*180.0/pi)+' '
        lin+=str(self.apt            )+' '
        lin+=str(self.phs_rf_typ     )+' '
        lin+=str(self.beta_s         )+' '
        lin+=str(self.T0             )+' '
        lin+=str(self.T1             )+' '
        lin+=str(self.T2             )

        return lin

    def get_madx(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name     +': RFCAVITY, L=0, '
        if self.name=='': lin='ELEM'+i+'_GAP'+': RFCAVITY, L=0, '
        lin+='FREQ='+str(self.freq                    )+', '
        lin+='VOLT='+str(self.E0TL                    )+', '
        lin+='LAG=' +str((0.5*pi-self.phs_rf)/(2.0*pi))+'; '  # phs_MADX = pi/2-phs_TW

        return lin

    def get_fluka(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name     +'  RFCAVITY  '
        if self.name=='': lin='ELEM'+i+'_GAP'+'  RFCAVITY  '
        lin+=str(self.L)+'  '+str(self.s)+'  '
        lin+='0.0  0.0  0.0  '
        lin+='CIRCLE  '+str(self.apt*1e-3)+'  0.0'

        return lin

    def get_mars(self):

        lin ='"RFCAVITY"  ""  ""  '+str(self.s)+'  '+str(self.L)+'  '
        lin+='0  0  0  0  0'

        return lin

class DTL_CEL(ELEM):
    '''
        - Note there are E0T0L (for TW) and E0TL (for other codes).
    '''
    def __init__(self,name,typ,para):

        ELEM.__init__(self,name,typ,para)

        # TW instances
        self.L         =float(para[0])*1e-3      # mm => m
        self.L_Q1      =float(para[1])*1e-3      # mm => m
        self.L_Q2      =float(para[2])*1e-3      # mm => m
        self.ds_gap    =float(para[3])*1e-3      # mm => m
        self.G_Q1      =float(para[4])
        self.G_Q2      =float(para[5])
        self.E0T0L     =float(para[6])*1e-6      # V => MV
        self.phs_rf    =float(para[7])*pi/180.0  # deg => rad
        self.apt       =float(para[8])
        self.phs_rf_typ=  int(para[9])

        # TW option instances
        try   : self.beta_s=float(para[10])
        except: self.beta_s=0.0
        try   : self.T0    =float(para[11])
        except: self.T0    =0.0
        try   : self.T1    =float(para[12])
        except: self.T1    =0.0
        try   : self.T2    =float(para[13])
        except: self.T2    =0.0

        # Option instances
        self.E0TL=self.E0T0L

    def update_gamma(self):

        # Save the gamma_ini (for MADX)
        self.gamma_ini=self.gamma

        # Update gamma (and E0TL)
        if self.beta_s!=0.0 and self.T0!=0.0:
            gamma_c   =self.gamma+0.5*self.E0TL/mass*cos(self.phs_rf)
            beta_c    =sqrt(1.0-1.0/gamma_c**2)
            k         =self.beta_s/beta_c
            self.E0TL*=(self.T0+self.T1*(k-1.0)+0.5*self.T2*(k-1.0)**2)/self.T0
        self.gamma+=self.E0TL/mass*cos(self.phs_rf)

    def get_tw(self):

        if self.name!='': lin=self.name+': '+self.typ+' '
        if self.name=='': lin=               self.typ+' '
        lin+=str(self.L*1e3          )+' '
        lin+=str(self.L_Q1*1e3       )+' '
        lin+=str(self.L_Q2*1e3       )+' '
        lin+=str(self.ds_gap*1e3     )+' '
        lin+=str(self.G_Q1           )+' '
        lin+=str(self.G_Q2           )+' '
        lin+=str(self.E0T0L*1e6      )+' '
        lin+=str(self.phs_rf*180.0/pi)+' '
        lin+=str(self.apt            )+' '
        lin+=str(self.phs_rf_typ     )+' '
        lin+=str(self.beta_s         )+' '
        lin+=str(self.T0             )+' '
        lin+=str(self.T1             )+' '
        lin+=str(self.T2             )

        return lin

    def get_madx(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name         +'_QUAD1: QUADRUPOLE, '
        if self.name=='': lin='ELEM'+i+'_DTLCELL'+'_QUAD1: QUADRUPOLE, '
        lin+='L=' +str(self.L_Q1                     )+', '
        lin+='K1='+str(self.G_Q1/Brho(self.gamma_ini))+'; \n'

        if self.name!='': lin+= self.name         +'_DRIFT1: DRIFT, '
        if self.name=='': lin+='ELEM'+i+'_DTLCELL'+'_DRIFT1: DRIFT, '
        lin+='L='+str(0.5*self.L-self.L_Q1-self.ds_gap)+';\n'

        if self.name!='': lin+= self.name         +'_GAP: RFCAVITY, L=0, '
        if self.name=='': lin+='ELEM'+i+'_DTLCELL'+'_GAP: RFCAVITY, L=0, '
        lin+='FREQ='+str(self.freq                    )+', '
        lin+='VOLT='+str(self.E0TL                    )+', '
        lin+='LAG=' +str((0.5*pi-self.phs_rf)/(2.0*pi))+'; \n'  # phs_MADX = pi/2-phs_TW

        if self.name!='': lin+= self.name         +'_DRIFT2: DRIFT, '
        if self.name=='': lin+='ELEM'+i+'_DTLCELL'+'_DRIFT2: DRIFT, '
        lin+='L='+str(0.5*self.L-self.L_Q2+self.ds_gap)+';\n'

        if self.name!='': lin+= self.name         +'_QUAD2: QUADRUPOLE, '
        if self.name=='': lin+='ELEM'+i+'_DTLCELL'+'_QUAD2: QUADRUPOLE, '
        lin+='L=' +str(self.L_Q2                 )+', '
        lin+='K1='+str(self.G_Q2/Brho(self.gamma))+'; '

        return lin

    def get_fluka(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        ## if self.name!='': lin= self.name         +'_QUAD1  QUADRUPOLE  '
        ## if self.name=='': lin='ELEM'+i+'_DTLCELL'+'_QUAD1  QUADRUPOLE  '
        ## lin+=str(self.L_Q1)+'  '+str(self.s-self.L+self.L_Q1)+'  '
        ## lin+='0.0  0.0  '+str(self.G_Q1*self.L_Q1/Brho(self.gamma_ini))+'  '
        ## lin+='CIRCLE  '+str(self.apt*1e-3)+'  0.0\n'

        ## if self.name!='': lin+= self.name         +'_DRIFT1  DRIFT  '
        ## if self.name=='': lin+='ELEM'+i+'_DTLCELL'+'_DRIFT1  DRIFT  '
        ## lin+=str(0.5*self.L-self.L_Q1-self.ds_gap)+'  '+str(self.s-0.5*self.L-self.ds_gap)+'  '
        ## lin+='0.0  0.0  0.0  '
        ## lin+='CIRCLE  '+str(self.apt*1e-3)+'  0.0\n'

        ## if self.name!='': lin+= self.name         +'_GAP  RFCAVITY   '
        ## if self.name=='': lin+='ELEM'+i+'_DTLCELL'+'_GAP  RFCAVITY   '
        ## lin+='0.0  '+str(self.s-0.5*self.L-self.ds_gap)+'  '
        ## lin+='0.0  0.0  0.0  '
        ## lin+='CIRCLE  '+str(self.apt*1e-3)+'  0.0\n'

        ## if self.name!='': lin+= self.name         +'_DRIFT2  DRIFT  '
        ## if self.name=='': lin+='ELEM'+i+'_DTLCELL'+'_DRIFT2  DRIFT  '
        ## lin+=str(0.5*self.L-self.L_Q2+self.ds_gap)+'  '+str(self.s-self.L_Q2)+'  '
        ## lin+='0.0  0.0  0.0  '
        ## lin+='CIRCLE  '+str(self.apt*1e-3)+'  0.0\n'

        ## if self.name!='': lin+= self.name         +'_QUAD2  QUADRUPOLE  '
        ## if self.name=='': lin+='ELEM'+i+'_DTLCELL'+'_QUAD2  QUADRUPOLE  '
        ## lin+=str(self.L_Q2)+'  '+str(self.s)+'  '
        ## lin+='0.0  0.0  '+str(self.G_Q2*self.L_Q2/Brho(self.gamma))+'  '
        ## lin+='CIRCLE  '+str(self.apt*1e-3)+'  0.0'

        if self.name!='': lin= self.name         +'  RFCAVITY  '
        if self.name=='': lin='ELEM'+i+'_DTLCELL'+'  RFCAVITY  '
        lin+=str(self.L)+'  '+str(self.s)+'  '
        lin+='0.0  0.0  0.0  '
        lin+='CIRCLE  '+str(self.apt*1e-3)+'  0.0'

        return lin

    def get_mars(self):

        lin ='"QUADRUPOLE"  ""  ""  '
        lin+=str(self.s-self.L+self.L_Q1)+'  '
        lin+=str(self.L_Q1)              +'  '
        lin+='0  '+str(self.G_Q1/Brho(self.gamma_ini))+'  0  0  0  \n'

        lin+='"DRIFT"  ""  ""  '
        lin+=str(self.s-0.5*self.L-self.ds_gap)   +'  '
        lin+=str(0.5*self.L-self.L_Q1-self.ds_gap)+'  '
        lin+='0  0  0  0  0  \n'

        lin+='"RFCAVITY"  ""  ""  '
        lin+=str(self.s-0.5*self.L-self.ds_gap)+'  '
        lin+='0  '
        lin+='0  0  0  0  0  \n'

        lin+='"DRIFT"  ""  ""  '
        lin+=str(self.s-self.L_Q2)                +'  '
        lin+=str(0.5*self.L-self.L_Q2+self.ds_gap)+'  '
        lin+='0  0  0  0  0  \n'

        lin+='"QUADRUPOLE"  ""  ""  '
        lin+=str(self.s)   +'  '
        lin+=str(self.L_Q2)+'  '
        lin+='0  '+str(self.G_Q2/Brho(self.gamma))+'  0  0  0  '

        return lin

#---- Passive elements

class BEND(ELEM):
    '''
        - For MADX, keeping the sbend+edge structure.
        - For MADX, if don't export edge, append instances of EDGE.
          (Otherwise, the matrix is wrong in MADX.)
        - For MARS/Fluka, assuming the rbend.
    '''
    def __init__(self,name,typ,para):

        ELEM.__init__(self,name,typ,para)

        # TW instances
        self.angle=float(para[0])*pi/180.0  # deg => rad
        self.rho  =float(para[1])*1e-3      # mm => m
        self.apt  =float(para[3])
        self.plane=      para[4]

        # Option instances
        if self.plane=='0': self.tilt=0.0
        if self.plane=='1': self.tilt=0.5*pi
        self.ds      =self.rho*abs(self.angle)
        self.L       =2.0*self.rho*sin(0.5*abs(self.angle))
        self.gap     =0.0  # Get this from EDGE IF not exporting edge for MADX
        self.ff_coef1=0.0  # Get this from EDGE IF not exporting edge for MADX
        self.ff_coef2=0.0  # Get this from EDGE IF not exporting edge for MADX

    def update_idx(self):

        # Updata idx, idx_elem, s
        self.idx     +=1
        self.idx_elem+=1
        self.s       +=self.ds  # Need an overwrite because of this

    def get_madx(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name      +': '
        if self.name=='': lin='ELEM'+i+'_BEND'+': '
        lin+='SBEND, L='+str(self.ds      )+', '  # sbend
        #lin+='RBEND, L='+str(self.L       )+', '  # rbend
        lin+='ANGLE='   +str(self.angle   )+', '
        lin+='TILT='    +str(self.tilt    )+', '
        lin+='HGAP='    +str(self.gap*5e-4)+', '
        lin+='FINT='    +str(self.ff_coef1)+'; '

        return lin

    def get_fluka(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name      +'  RBEND  '
        if self.name=='': lin='ELEM'+i+'_BEND'+'  RBEND  '
        lin+=str(self.ds)+'  '+str(self.s)+'  '
        lin+=str(self.tilt)+'  '+str(self.angle)+'  0.0  '
        lin+='CIRCLE  '+str(self.apt*1e-3)+'  0.0'

        return lin

    def get_mars(self):

        lin ='"RBEND"  ""  ""  '+str(self.s)+'  '+str(self.L)+'  '
        lin+=str(self.angle)+'  0  0  0  '+str(self.tilt)

        return lin

class EDGE(ELEM):
    '''
        - For MARS, assuming rbend and the export is commented out.
    '''
    def __init__(self,name,typ,para):

        ELEM.__init__(self,name,typ,para)

        # TW instances
        self.angle   =float(para[0])*pi/180.0  # deg => rad
        self.rho     =float(para[1])*1e-3      # mm => m
        self.gap     =float(para[2])
        self.ff_coef1=float(para[3])
        self.ff_coef2=float(para[4])
        self.apt     =float(para[5])
        self.plane   =      para[6]

        # Option instances
        if self.plane=='0': self.tilt=0.0
        if self.plane=='1': self.tilt=0.5*pi

    def get_madx(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name      +': DIPEDGE, '
        if self.name=='': lin='ELEM'+i+'_EDGE'+': DIPEDGE, '
        lin+='E1='  +str(self.angle)   +', '  # sbend
        #lin+='E1='  +'0'               +', '  # rbend
        lin+='H='   +str(1.0/self.rho) +', '
        lin+='TILT='+str(self.tilt    )+', '
        lin+='HGAP='+str(self.gap*5e-4)+', '
        lin+='FINT='+str(self.ff_coef1)+'; '

        return lin

class APERTURE(ELEM):
    '''
        - For MADX, tried "collimator" but didn't work ...?
    '''
    def __init__(self,name,typ,para):

        ELEM.__init__(self,name,typ,para)

        # TW instances
        self.apt    =float(para[0])
        self.apty   =float(para[1])
        self.typ_apt=      para[2]

    def get_madx(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name     +': DRIFT, L=0;'
        if self.name=='': lin='ELEM'+i+'_APT'+': DRIFT, L=0;'

        ## if self.name!='': lin= self.name     +': collimator, l=0, '
        ## if self.name=='': lin='ELEM'+i+'_apt'+': collimator, l=0, '
        ## if self.typ_apt=='0':
        ##     lin+='apertype=rectangle, aperture={'+str(self.apt*1e-3)+','+str(self.apty*1e-3)+'};'
        ## if self.typ_apt=='1':
        ##     lin+='apertype=CIRCLE, aperture={'+str(self.apt*1e-3)+'};'

        return lin

    def get_fluka(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name     +'  DRIFT  '
        if self.name=='': lin='ELEM'+i+'_APT'+'  DRIFT  '
        lin+=str(self.L)+'  '+str(self.s)+'  '
        lin+='0.0  0.0  0.0  '
        if self.typ_apt=='0': lin+='RECTANGLE  '+str(self.apt*1e-3)+'  '+str(self.apty*1e-3)
        if self.typ_apt=='1': lin+='CIRCLE  '+str(self.apt*1e-3)+'  0.0'

        return lin

    def get_mars(self):

        lin ='"drift"  ""  ""  '+str(self.s)+'  '+str(self.L)+'  '
        lin+='0  0  0  0  0'

        return lin

class DIAG_POSITION(ELEM):
    '''
    '''
    def __init__(self,name,typ,para):

        ELEM.__init__(self,name,typ,para)

    def get_madx(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name     +': MONITOR, L=0;'
        if self.name=='': lin='ELEM'+i+'_BPM'+': MONITOR, L=0;'

        return lin

    def get_fluka(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name     +'  MONITOR  '
        if self.name=='': lin='ELEM'+i+'_BPM'+'  MONITOR  '
        lin+=str(self.L)+'  '+str(self.s)+'  '
        lin+='0.0  0.0  0.0  '
        lin+='CIRCLE  '+str(self.apt*1e-3)+'  0.0'

        return lin

    def get_mars(self):

        lin ='"MONITOR"  ""  ""  '+str(self.s)+'  '+str(self.L)+'  '
        lin+='0  0  0  0  0'

        return lin

class DIAG(ELEM):
    '''
    '''
    def __init__(self,name,typ,para):

        ELEM.__init__(self,name,typ,para)

    def get_madx(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name      +': INSTRUMENT, L=0;'
        if self.name=='': lin='ELEM'+i+'_DIAG'+': INSTRUMENT, L=0;'

        return lin

    def get_fluka(self):

        i='0'*(3-int(log10(self.idx_elem+1)))+str(self.idx_elem+1)

        if self.name!='': lin= self.name      +'  INSTRUMENT  '
        if self.name=='': lin='ELEM'+i+'_DIAG'+'  INSTRUMENT  '
        lin+=str(self.L)+'  '+str(self.s)+'  '
        lin+='0.0  0.0  0.0  '
        lin+='CIRCLE  '+str(self.apt*1e-3)+'  0.0'

        return lin

    def get_mars(self):

        lin ='"INSTRUMENT"  ""  ""  '+str(self.s)+'  '+str(self.L)+'  '
        lin+='0  0  0  0  0'

        return lin

#---- Active commands

class STEERER(COMM):
    '''
        - Only magnetic steerers supported.
        - Only the ESS type non-linearity supported.
    '''
    def __init__(self,name,typ,para):

        COMM.__init__(self,name,typ,para)

        # TW instances
        self.Bx=float(para[0])
        self.By=float(para[1])

        # TW option instances
        try   : self.max   =float(para[2])
        except: self.max   =0.0
        try   : self.nonlin=float(para[5])
        except: self.nonlin=0.0

    def get_tw(self):

        if self.name!='': lin=self.name+': '+self.typ+' '
        if self.name=='': lin=               self.typ+' '
        lin+=str(self.Bx    )+' '
        lin+=str(self.By    )+' '
        lin+=str(self.max   )+' '
        lin+='0'             +' '
        lin+='0'             +' '
        lin+=str(self.nonlin)

        return lin

class CHOPPER(COMM):
    '''
    '''
    def __init__(self,name,typ,para):

        COMM.__init__(self,name,typ,para)

        # TW instances
        self.Nelem=  int(para[0])
        self.volt =float(para[1])
        self.gap  =float(para[2])*2.0
        self.d_pos=float(para[3])
        self.plane=      para[4]

    def get_tw(self):

        if self.name!='': lin=self.name+': '+self.typ+' '
        if self.name=='': lin=               self.typ+' '
        lin+=str(self.Nelem  )+' '
        lin+=str(self.volt   )+' '
        lin+=str(self.gap*0.5)+' '
        lin+=str(self.d_pos  )+' '
        lin+=    self.plane

        return lin

#---- Passive commands

class FREQ(COMM):
    '''
    '''
    def __init__(self,name,typ,para):

        COMM.__init__(self,name,typ,para)

        # TW instances
        self.freq_new=float(para[0])

    def update_idx(self):

        self.idx +=1
        self.freq =self.freq_new

class MARKER(COMM):
    '''
    '''
    def __init__(self,name,typ,para):

        COMM.__init__(self,name,typ,para)

    def get_madx(self):

        if self.name!='': lin=self.name   +': MARKER;'
        if self.name=='': lin=self.para[0]+': MARKER;'

        return lin

    def get_fluka(self):

        if self.name!='': lin=self.name   +'  MARKER  '
        if self.name=='': lin=self.para[0]+'  MARKER  '
        lin+='0.0  '+str(self.s)

        return lin

    ## def get_mars(self):

    ##     lin ='"marker"  ""  ""  '+str(self.s)+'  0.0  '
    ##     lin+='0  0  0  0  0'

    ##     return lin

#---- No longer needed ??

## class ADJUST(COMM):
##     '''
##     '''
##     def __init__(self,name,typ,para):

##         COMM.__init__(self,name,typ,para)

##         # TW instances
##         self.fam=    para[0]
##         self.var=int(para[1])

##         # TW option instances
##         try   : self.link=self.para[2]
##         except: self.link='0'
##         try   : self.min =float(self.para[3])
##         except: self.min =0.0
##         try   : self.max =float(self.para[4])
##         except: self.max =0.0
##         try   : self.ini =float(self.para[5])
##         except: self.ini =0.0
